module NScript

  class Scope

    attr_reader :parent, :expressions, :function, :variables, :temp_variable

    def initialize(parent, expressions, function)
      @parent, @expressions, @function = parent, expressions, function
      @variables = {}
      @temp_variable = @parent ? @parent.temp_variable.dup : '__a'
    end

    def find(name, remote=false)
      found = check(name)
      return found if found || remote
      @variables[name.to_sym] = :var
      found
    end

    def parameter(name)
      @variables[name.to_sym] = :param
    end

    def check(name)
      return true if @variables[name.to_sym]
      !!(@parent && @parent.check(name))
    end

    def reset(name)
      @variables[name.to_sym] = false
    end

    def free_variable
      @temp_variable.succ! while check(@temp_variable)
      @variables[@temp_variable.to_sym] = :var
      Value.new(@temp_variable.dup)
    end

    def assign(name, value, top=false)
      return @parent.assign(name, value, top) if top && @parent
      @variables[name.to_sym] = Value.new(value)
    end

    def declarations?(body)
      !declared_variables.empty? && body == @expressions
    end

    def assignments?(body)
      !assigned_variables.empty? && body == @expressions
    end

    def declared_variables
      @variables.select {|k, v| v == :var }.map {|pair| pair[0].to_s }.sort
    end

    def assigned_variables
      @variables.select {|k, v| v.is_a?(Value) }.sort_by {|pair| pair[0].to_s }
    end

    def compiled_declarations
      declared_variables.join(', ')
    end

    def compiled_assignments
      assigned_variables.map {|name, val| "#{name} = #{val}"}.join(', ')
    end

    def inspect
      "<Scope:#{__id__} #{@variables.inspect}>"
    end

  end

end
